#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Module to load OPL instrument files.
#

import os
import sys
import sbi_file
import a2i_file

# Check the specified instrument data is OPL2-compatible and does not
# use any OPL3 features. Returns an error message, or 'None' if data
# is valid.


def check_opl2(filename, data):
    def opl2_warning(message):
        print("%s: %s" % (filename, message), file=sys.stderr)

    # CHA,B control stereo, but are ignored on OPL2, so it's no problem:
    # if (data["feedback_fm"] & 0xf0) != 0:
    # 	opl2_warning("Cannot use CHA,B,C,D: %02x" % data["feedback_fm"])

    if data["m_waveform"] > 3:
        opl2_warning(
            "Modulator uses waveform %i: only 0-3 supported"
            % data["m_waveform"]
        )
    if data["c_waveform"] > 3:
        opl2_warning(
            "Carrier uses waveform %i: only 0-3 supported" % data["c_waveform"]
        )


def load_instrument(filename):

    # As a hack, a literal dictionary of the values can be specified
    # in place of a filename.

    if isinstance(filename, dict):
        return filename

    filename = os.path.join("instruments", filename)

    if filename.endswith(".a2i"):
        result = a2i_file.read(filename)
    elif filename.endswith(".sbi"):
        result = sbi_file.read(filename)
    else:
        raise Exception("Unknown instrument file type: '%s'" % filename)

    check_opl2(filename, result)

    return result


class Instrument:
    def __init__(self, file1, file2=None, off1=0, off2=0, note=None):
        self.voice1 = load_instrument(file1)

        if file2 is not None:
            self.voice2 = load_instrument(file2)
        else:
            self.voice2 = None

        self.fixed_note = note
        self.offset1 = off1
        self.offset2 = off2


NullInstrument = Instrument("dummy.sbi")

if __name__ == "__main__":
    for filename in sys.argv[1:]:
        Instrument(filename)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Constants for MIDI notes.
#
# For example:
#    F# in Octave 3:      O3.Fs
#    C# in Octave -2:     On2.Cs
#    D-flat in Octave 1:  O1.Db
#    D in Octave 0:       O0.D
#    E in Octave 2:       O2.E


class Octave:
    def __init__(self, base):
        self.C = base
        self.Cs = base + 1
        self.Db = base + 1
        self.D = base + 2
        self.Ds = base + 3
        self.Eb = base + 3
        self.E = base + 4
        self.F = base + 5
        self.Fs = base + 6
        self.Gb = base + 6
        self.G = base + 7
        self.Gs = base + 8
        self.Ab = base + 8
        self.A = base + 9
        self.As = base + 10
        self.Bb = base + 10
        self.B = base + 11


On5 = Octave(0)  # Octave -5
On4 = Octave(12)  # Octave -4
On3 = Octave(24)  # Octave -3
On2 = Octave(36)  # Octave -2
On1 = Octave(48)  # Octave -1
O0 = Octave(60)  # Octave 0
O1 = Octave(72)  # Octave 1
O2 = Octave(84)  # Octave 2
O3 = Octave(96)  # Octave 3
O4 = Octave(108)  # Octave 4
O5 = Octave(120)  # Octave 5

# Given a MIDI note number, return a note definition in terms of the
# constants above.


def def_for_note(note):
    OCTAVES = [
        "On5",
        "On4",
        "On3",
        "On2",
        "On1",
        "O0",
        "O1",
        "O2",
        "O3",
        "O4",
        "O5",
    ]
    NOTES = ["C", "Cs", "D", "Ds", "E", "F", "Fs", "G", "Gs", "A", "As", "B"]

    return "%s.%s" % (OCTAVES[note // 12], NOTES[note % 12])



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Module to load A2I (AdlibTracker 2) instrument files.
#

import sys
import struct

HEADER_STRING = "_A2ins_"


class BitReader:
    def __init__(self, data):
        self.data = data
        self.index = 0
        self.byte = 0
        self.byte_bit = 0

    def read_byte(self):
        if self.index >= len(self.data):
            raise IndexError(
                "Reached end of decompress stream "
                + "(%i bytes)" % len(self.data)
            )
        result, = struct.unpack("B", self.data[self.index : self.index + 1])
        self.index += 1
        return result

    def read_bit(self):
        if self.byte_bit <= 0:
            self.byte = self.read_byte()
            self.byte_bit = 7
        else:
            self.byte_bit -= 1

        if (self.byte & (1 << self.byte_bit)) != 0:
            result = 1
        else:
            result = 0

        return result

    def read_bits(self, n):
        result = 0

        for i in range(n):
            result = (result << 1) + self.read_bit()

        return result


def read_gamma(reader):
    result = 1

    while True:
        result = (result << 1) | reader.read_bit()

        if reader.read_bit() == 0:
            break

    return result


def decompress(data, data_len):
    reader = BitReader(data)
    result = []
    lwm = 0
    last_offset = 0

    # First byte is an implied straight copy.
    result.append(reader.read_byte())

    while True:
        if reader.read_bit():
            if reader.read_bit():
                if reader.read_bit():
                    # 111 = Copy byte from history,
                    # up to 15 bytes back.
                    # print "111 copy"

                    offset = reader.read_bits(4)

                    if offset == 0:
                        result.append(0)
                    else:
                        b = result[len(result) - offset]
                        result.append(b)

                    lwm = 0
                else:
                    # print "110 copy"
                    # 110 = Copy 2-3 bytes from
                    # further back in history

                    offset = reader.read_byte()
                    count = 2 + (offset & 0x01)
                    offset = offset >> 1

                    if offset == 0:
                        break

                    index = len(result) - offset
                    for i in range(count):
                        result += result[index : index + 1]
                        index += 1

                    last_offset = offset
                    lwm = 1
            else:
                # 10 = Copy from further away...

                offset = read_gamma(reader)

                if lwm == 0 and offset == 2:
                    # print "10 copy type 1"
                    count = read_gamma(reader)
                    index = len(result) - last_offset
                    for i in range(count):
                        result += result[index : index + 1]
                        index += 1
                else:
                    # print "10 copy type 2"
                    if lwm == 0:
                        offset -= 3
                    else:
                        offset -= 2

                    offset = offset * 256 + reader.read_byte()

                    count = read_gamma(reader)

                    if offset >= 32000:
                        count += 1
                    if offset >= 1280:
                        count += 1
                    if offset < 128:
                        count += 2

                    index = len(result) - offset
                    for i in range(count):
                        result += result[index : index + 1]
                        index += 1

                    last_offset = offset

                lwm = 1

        else:
            # print "Single byte output"

            # 0 = Straight-through byte copy.
            result.append(reader.read_byte())
            lwm = 0

        # print "len: %i" % len(result)

    return struct.pack("%iB" % len(result), *result)


FIELDS = [
    "m_am_vibrato_eg",
    "c_am_vibrato_eg",
    "m_ksl_volume",
    "c_ksl_volume",
    "m_attack_decay",
    "c_attack_decay",
    "m_sustain_release",
    "c_sustain_release",
    "m_waveform",
    "c_waveform",
    "feedback_fm",
    "panning",
    "finetune",
    "voice_type",
]


def decode_type_9(data):
    compressed_len, = struct.unpack("<H", data[0:2])
    compressed_data = data[2 : 2 + compressed_len]
    decompressed_data = decompress(compressed_data, compressed_len)

    instr_data = {}

    # Decode main fields:

    for i in range(len(FIELDS)):
        instr_data[FIELDS[i]], = struct.unpack(
            "B", decompressed_data[i : i + 1]
        )

    # Decode instrument name

    ps = decompressed_data[14:]
    instr_name, = struct.unpack("%ip" % len(ps), ps)
    instr_data["name"] = instr_name.decode("ascii")

    return instr_data


def read(filename):
    with open(filename, "rb") as f:
        data = f.read()

    hdrstr, crc, filever = struct.unpack("<7sHB", data[0:10])

    hdrstr = hdrstr.decode("ascii")

    if hdrstr.lower() != HEADER_STRING.lower():
        raise Exception("Wrong file header ID string")

    # TODO: CRC?

    if filever == 9:
        return decode_type_9(data[10:])
    else:
        raise Exception("Unsupported file version: %i" % filever)


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        print(filename)
        print(read(filename))



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Script to gather statistics from MIDI files about instrument usage.
# Generates stats.py which is used as input for the ultramid.ini
# generator script.
#
# This script requires python-midi; see here:
#   https://github.com/vishnubob/python-midi
#

import midi
import sys


def get_instr_stats(filename):
    """Get a set of instruments used by the specified MIDI file."""
    result = set()
    midfile = midi.read_midifile(filename)

    for track in midfile:
        for event in track:
            if (
                isinstance(event, midi.ProgramChangeEvent)
                and event.channel != 9
            ):
                instr = event.data[0]
                result.add(instr)
            # Percussion:
            if isinstance(event, midi.NoteOnEvent) and event.channel == 9:
                instr = event.data[0] + 128
                result.add(instr)

    return result


total_stats = [0] * 217

for filename in sys.argv[1:]:
    print("Processing %s" % filename)
    stats = get_instr_stats(filename)
    print(sorted(stats))
    for instrument in stats:
        total_stats[instrument] += 1

with open("stats.py", "w") as f:
    f.write("# Instrument stats, autogenerated by gather_stats.py\n\n")
    f.write("INSTRUMENT_STATS = [\n\t")

    for index, stat in enumerate(total_stats):
        f.write("% 5i," % stat)
        if (index % 10) == 9:
            f.write("\n\t")

    f.write("\n]\n")



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Module for interacting with Doom GENMIDI lumps.
#

from instrument import Instrument, NullInstrument
import struct
import sys

GENMIDI_HEADER = "#OPL_II#"
NUM_INSTRUMENTS = 175
INSTR_DATA_LEN = 36
INSTR_NAME_LEN = 32

FLAG_FIXED_PITCH = 0x0001
FLAG_TWO_VOICE = 0x0004

KSL_MASK = 0xC0
VOLUME_MASK = 0x3F

# Order of fields in GENMIDI data structures.

GENMIDI_FIELDS = [
    "m_am_vibrato_eg",
    "m_attack_decay",
    "m_sustain_release",
    "m_waveform",
    "m_ksl",
    "m_volume",
    "feedback_fm",
    "c_am_vibrato_eg",
    "c_attack_decay",
    "c_sustain_release",
    "c_waveform",
    "c_ksl",
    "c_volume",
    "null",
    "note_offset",
]

# Encode a single voice of an instrument to binary.


def encode_voice(data, offset):
    result = dict(data)

    result["m_ksl"] = data["m_ksl_volume"] & KSL_MASK
    result["m_volume"] = data["m_ksl_volume"] & VOLUME_MASK
    result["c_ksl"] = data["c_ksl_volume"] & KSL_MASK
    result["c_volume"] = data["c_ksl_volume"] & VOLUME_MASK

    result["null"] = 0
    result["note_offset"] = offset

    return struct.pack(
        "<BBBBBBBBBBBBBBh", *map(lambda key: result[key], GENMIDI_FIELDS)
    )


# Encode an instrument to binary.


def encode_instrument(instrument):
    flags = 0

    instr1_data = encode_voice(instrument.voice1, instrument.offset1)

    if instrument.voice2 is not None:
        flags |= FLAG_TWO_VOICE
        instr2_data = encode_voice(instrument.voice2, instrument.offset2)
    else:
        instr2_data = encode_voice(NullInstrument.voice1, 0)

    if instrument.fixed_note is not None:
        flags |= FLAG_FIXED_PITCH
        fixed_note = instrument.fixed_note
    else:
        fixed_note = 0

    header = struct.pack("<hBB", flags, 128, fixed_note)

    return header + instr1_data + instr2_data


def encode_instruments(instruments):
    result = []

    for instrument in instruments:
        result.append(encode_instrument(instrument))

    return b"".join(result)


def encode_instrument_names(instruments):
    result = []

    for instrument in instruments:
        instr_name = instrument.voice1["name"].encode("ascii")
        result.append(struct.pack("32s", instr_name))

    return b"".join(result)


def write(filename, instruments):
    header = struct.pack(
        "%is" % len(GENMIDI_HEADER), GENMIDI_HEADER.encode("ascii")
    )

    with open(filename, "wb") as f:
        f.write(header)
        f.write(encode_instruments(instruments))
        f.write(encode_instrument_names(instruments))


def decode_voice(data, name):

    fields = struct.unpack("<BBBBBBBBBBBBBBh", data)

    result = {}
    for i in range(len(GENMIDI_FIELDS)):
        result[GENMIDI_FIELDS[i]] = fields[i]

    result["m_ksl_volume"] = result["m_ksl"] | result["m_volume"]
    result["c_ksl_volume"] = result["c_ksl"] | result["c_volume"]
    result["name"] = name.decode("ascii").rstrip("\0")

    return result


def decode_instrument(data, name):
    flags, finetune, fixed_note = struct.unpack("<hBB", data[0:4])

    voice1 = decode_voice(data[4:20], name)
    offset1 = voice1["note_offset"]

    # Second voice?

    if (flags & FLAG_TWO_VOICE) != 0:
        voice2 = decode_voice(data[20:], name)
        offset2 = voice2["note_offset"]
    else:
        voice2 = None
        offset2 = 0

    # Null out fixed_note if the fixed pitch flag isn't set:

    if (flags & FLAG_FIXED_PITCH) == 0:
        fixed_note = None

    return Instrument(
        voice1, voice2, off1=offset1, off2=offset2, note=fixed_note
    )


def read(filename):
    with open(filename, "rb") as f:
        data = f.read()

    # Check header:

    header = data[0 : len(GENMIDI_HEADER)]
    if header.decode("ascii") != GENMIDI_HEADER:
        raise Exception("Incorrect header for GENMIDI lump")

    body = data[len(GENMIDI_HEADER) :]
    instr_data = body[0 : NUM_INSTRUMENTS * INSTR_DATA_LEN]
    instr_names = body[NUM_INSTRUMENTS * INSTR_DATA_LEN :]
    result = []

    for i in range(NUM_INSTRUMENTS):
        data = instr_data[i * INSTR_DATA_LEN : (i + 1) * INSTR_DATA_LEN]
        name = instr_names[i * INSTR_NAME_LEN : (i + 1) * INSTR_NAME_LEN]

        result.append(decode_instrument(data, name))

    return result


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        instruments = read(filename)
        for i in range(len(instruments)):
            instrument = instruments[i]
            fixed_note = instrument.fixed_note

            if fixed_note is not None:
                print("%i (fixed note: %i):" % (i, fixed_note))
            else:
                print("%i:" % i)

            print("\tVoice 1: %s" % instrument.voice1)
            if instrument.voice2 is not None:
                print("\tVoice 2: %s" % instrument.voice2)



# SPDX-License-Identifier: BSD-3-Clause
#
# Config file for GUS config generator.
#

# Names of GUS instrument files to use for playing MIDI.
# These are the names of the .pat files in the \ultrasnd\midi directory
# that are loaded into the card.

GUS_INSTR_PATCHES = {
    0: "acpiano",  # #001 - Acoustic Grand Piano
    1: "britepno",  # #002 - Bright Acoustic Piano
    2: "synpiano",  # #003 - Electric Grand Piano
    3: "honky",  # #004 - Honky-tonk Piano
    4: "epiano1",  # #005 - Electric Piano 1
    5: "epiano2",  # #006 - Electric Piano 2
    6: "hrpschrd",  # #007 - Harpsichord
    7: "clavinet",  # #008 - Clavi
    8: "celeste",  # #009 - Celesta
    9: "glocken",  # #010 - Glockenspiel
    10: "musicbox",  # #011 - Music Box
    11: "vibes",  # #012 - Vibraphone
    12: "marimba",  # #013 - Marimba
    13: "xylophon",  # #014 - Xylophone
    14: "tubebell",  # #015 - Tubular Bells
    15: "santur",  # #016 - Dulcimer
    16: "homeorg",  # #017 - Drawbar Organ
    17: "percorg",  # #018 - Percussive Organ
    18: "rockorg",  # #019 - Rock Organ
    19: "church",  # #020 - Church Organ
    20: "reedorg",  # #021 - Reed Organ
    21: "accordn",  # #022 - Accordion
    22: "harmonca",  # #023 - Harmonica
    23: "concrtna",  # #024 - Tango Accordion
    24: "nyguitar",  # #025 - Acoustic Guitar (nylon)
    25: "acguitar",  # #026 - Acoustic Guitar (steel)
    26: "jazzgtr",  # #027 - Electric Guitar (jazz)
    27: "cleangtr",  # #028 - Electric Guitar (clean)
    28: "mutegtr",  # #029 - Electric Guitar (muted)
    29: "odguitar",  # #030 - Overdriven Guitar
    30: "distgtr",  # #031 - Distortion Guitar
    31: "gtrharm",  # #032 - Guitar harmonics
    32: "acbass",  # #033 - Acoustic Bass
    33: "fngrbass",  # #034 - Electric Bass (finger)
    34: "pickbass",  # #035 - Electric Bass (pick)
    35: "fretless",  # #036 - Fretless Bass
    36: "slapbas1",  # #037 - Slap Bass 1
    37: "slapbas2",  # #038 - Slap Bass 2
    38: "synbass1",  # #039 - Synth Bass 1
    39: "synbass2",  # #040 - Synth Bass 2
    40: "violin",  # #041 - Violin
    41: "viola",  # #042 - Viola
    42: "cello",  # #043 - Cello
    43: "contraba",  # #044 - Contrabass
    44: "tremstr",  # #045 - Tremolo Strings
    45: "pizzcato",  # #046 - Pizzicato Strings
    46: "harp",  # #047 - Orchestral Harp
    47: "timpani",  # #048 - Timpani
    48: "marcato",  # #049 - String Ensemble 1
    49: "slowstr",  # #050 - String Ensemble 2
    50: "synstr1",  # #051 - SynthStrings 1
    51: "synstr2",  # #052 - SynthStrings 2
    52: "choir",  # #053 - Choir Aahs
    53: "doo",  # #054 - Voice Oohs
    54: "voices",  # #055 - Synth Voice
    55: "orchhit",  # #056 - Orchestra Hit
    56: "trumpet",  # #057 - Trumpet
    57: "trombone",  # #058 - Trombone
    58: "tuba",  # #059 - Tuba
    59: "mutetrum",  # #060 - Muted Trumpet
    60: "frenchrn",  # #061 - French Horn
    61: "hitbrass",  # #062 - Brass Section
    62: "synbras1",  # #063 - SynthBrass 1
    63: "synbras2",  # #064 - SynthBrass 2
    64: "sprnosax",  # #065 - Soprano Sax
    65: "altosax",  # #066 - Alto Sax
    66: "tenorsax",  # #067 - Tenor Sax
    67: "barisax",  # #068 - Baritone Sax
    68: "oboe",  # #069 - Oboe
    69: "englhorn",  # #070 - English Horn
    70: "bassoon",  # #071 - Bassoon
    71: "clarinet",  # #072 - Clarinet
    72: "piccolo",  # #073 - Piccolo
    73: "flute",  # #074 - Flute
    74: "recorder",  # #075 - Recorder
    75: "woodflut",  # #076 - Pan Flute
    76: "bottle",  # #077 - Blown Bottle
    77: "shakazul",  # #078 - Shakuhachi
    78: "whistle",  # #079 - Whistle
    79: "ocarina",  # #080 - Ocarina
    80: "sqrwave",  # #081 - Lead 1 (square)
    81: "sawwave",  # #082 - Lead 2 (sawtooth)
    82: "calliope",  # #083 - Lead 3 (calliope)
    83: "chiflead",  # #084 - Lead 4 (chiff)
    84: "charang",  # #085 - Lead 5 (charang)
    85: "voxlead",  # #086 - Lead 6 (voice)
    86: "lead5th",  # #087 - Lead 7 (fifths)
    87: "basslead",  # #088 - Lead 8 (bass + lead)
    88: "fantasia",  # #089 - Pad 1 (new age)
    89: "warmpad",  # #090 - Pad 2 (warm)
    90: "polysyn",  # #091 - Pad 3 (polysynth)
    91: "ghostie",  # #092 - Pad 4 (choir)
    92: "bowglass",  # #093 - Pad 5 (bowed)
    93: "metalpad",  # #094 - Pad 6 (metallic)
    94: "halopad",  # #095 - Pad 7 (halo)
    95: "sweeper",  # #096 - Pad 8 (sweep)
    96: "aurora",  # #097 - FX 1 (rain)
    97: "soundtrk",  # #098 - FX 2 (soundtrack)
    98: "crystal",  # #099 - FX 3 (crystal)
    99: "atmosphr",  # #100 - FX 4 (atmosphere)
    100: "freshair",  # #101 - FX 5 (brightness)
    101: "unicorn",  # #102 - FX 6 (goblins)
    102: "echovox",  # #103 - FX 7 (echoes)
    103: "startrak",  # #104 - FX 8 (sci-fi)
    104: "sitar",  # #105 - Sitar
    105: "banjo",  # #106 - Banjo
    106: "shamisen",  # #107 - Shamisen
    107: "koto",  # #108 - Koto
    108: "kalimba",  # #109 - Kalimba
    109: "bagpipes",  # #110 - Bag pipe
    110: "fiddle",  # #111 - Fiddle
    111: "shannai",  # #112 - Shanai
    112: "carillon",  # #113 - Tinkle Bell
    113: "agogo",  # #114 - Agogo
    114: "steeldrm",  # #115 - Steel Drums
    115: "woodblk",  # #116 - Woodblock
    116: "taiko",  # #117 - Taiko Drum
    117: "toms",  # #118 - Melodic Tom
    118: "syntom",  # #119 - Synth Drum
    119: "revcym",  # #120 - Reverse Cymbal
    120: "fx-fret",  # #121 - Guitar Fret Noise
    121: "fx-blow",  # #122 - Breath Noise
    122: "seashore",  # #123 - Seashore
    123: "jungle",  # #124 - Bird Tweet
    124: "telephon",  # #125 - Telephone Ring
    125: "helicptr",  # #126 - Helicopter
    126: "applause",  # #127 - Applause
    127: "pistol",  # #128 - Gunshot
    128: "blank",
    163: "kick1",  # #35 Acoustic Bass Drum
    164: "kick2",  # #36 Bass Drum 1
    165: "stickrim",  # #37 Side Stick
    166: "snare1",  # #38 Acoustic Snare
    167: "claps",  # #39 Hand Clap
    168: "snare2",  # #40 Electric Snare
    169: "tomlo2",  # #41 Low Floor Tom
    170: "hihatcl",  # #42 Closed Hi Hat
    171: "tomlo1",  # #43 High Floor Tom
    172: "hihatpd",  # #44 Pedal Hi-Hat
    173: "tommid2",  # #45 Low Tom
    174: "hihatop",  # #46 Open Hi-Hat
    175: "tommid1",  # #47 Low-Mid Tom
    176: "tomhi2",  # #48 Hi-Mid Tom
    177: "cymcrsh1",  # #49 Crash Cymbal 1
    178: "tomhi1",  # #50 High Tom
    179: "cymride1",  # #51 Ride Cymbal 1
    180: "cymchina",  # #52 Chinese Cymbal
    181: "cymbell",  # #53 Ride Bell
    182: "tamborin",  # #54 Tambourine
    183: "cymsplsh",  # #55 Splash Cymbal
    184: "cowbell",  # #56 Cowbell
    185: "cymcrsh2",  # #57 Crash Cymbal 2
    186: "vibslap",  # #58 Vibraslap
    187: "cymride2",  # #59 Ride Cymbal 2
    188: "bongohi",  # #60 Hi Bongo
    189: "bongolo",  # #61 Low Bongo
    190: "congahi1",  # #62 Mute Hi Conga
    191: "congahi2",  # #63 Open Hi Conga
    192: "congalo",  # #64 Low Conga
    193: "timbaleh",  # #65 High Timbale
    194: "timbalel",  # #66 Low Timbale
    195: "agogohi",  # #67 High Agogo
    196: "agogolo",  # #68 Low Agogo
    197: "cabasa",  # #69 Cabasa
    198: "maracas",  # #70 Maracas
    199: "whistle1",  # #71 Short Whistle
    200: "whistle2",  # #72 Long Whistle
    201: "guiro1",  # #73 Short Guiro
    202: "guiro2",  # #74 Long Guiro
    203: "clave",  # #75 Claves
    204: "woodblk1",  # #76 Hi Wood Block
    205: "woodblk2",  # #77 Low Wood Block
    206: "cuica1",  # #78 Mute Cuica
    207: "cuica2",  # #79 Open Cuica
    208: "triangl1",  # #80 Mute Triangle
    209: "triangl2",  # #81 Open Triangle
}

# These are the data sizes of the patch files distributed with the
# GUS drivers. These are used to calculate the size in RAM of the
# generated patch sets to check that they are within the limits.
# and check it is within the limit.

PATCH_FILE_SIZES = {
    "acbass": 5248,
    "accordn": 9616,
    "acguitar": 26080,
    "acpiano": 32256,
    "agogo": 13696,
    "agogohi": 3488,
    "agogolo": 3488,
    "altosax": 5616,
    "applause": 30064,
    "atmosphr": 31360,
    "aurora": 31088,
    "bagpipes": 7760,
    "banjo": 32016,
    "barisax": 10544,
    "basslead": 26496,
    "bassoon": 8000,
    "belltree": 31888,
    "blank": 1520,
    "bongohi": 3456,
    "bongolo": 4448,
    "bottle": 12368,
    "bowglass": 24688,
    "britepno": 36000,
    "cabasa": 8448,
    "calliope": 22992,
    "carillon": 5888,
    "castinet": 6016,
    "celeste": 9936,
    "cello": 9120,
    "charang": 45056,
    "chiflead": 31536,
    "choir": 22480,
    "church": 2144,
    "claps": 5696,
    "clarinet": 9184,
    "clave": 2352,
    "clavinet": 1440,
    "cleangtr": 22768,
    "concrtna": 8784,
    "congahi1": 4224,
    "congahi2": 4704,
    "congalo": 4704,
    "contraba": 4704,
    "cowbell": 3168,
    "crystal": 30224,
    "cuica1": 9344,
    "cuica2": 12848,
    "cymbell": 17248,
    "cymchina": 24112,
    "cymcrsh1": 31520,
    "cymcrsh2": 31040,
    "cymride1": 17664,
    "cymride2": 17664,
    "cymsplsh": 31520,
    "distgtr": 18848,
    "doo": 8464,
    "echovox": 14976,
    "englhorn": 12096,
    "epiano1": 7344,
    "epiano2": 21936,
    "fantasia": 23456,
    "fiddle": 5904,
    "flute": 6032,
    "fngrbass": 9744,
    "frenchrn": 14128,
    "freshair": 28992,
    "fretless": 2640,
    "fx-blow": 28688,
    "fx-fret": 13648,
    "ghostie": 31488,
    "glocken": 5184,
    "gtrharm": 4928,
    "guiro1": 4128,
    "guiro2": 9248,
    "halopad": 29984,
    "harmonca": 7408,
    "harp": 11728,
    "helicptr": 25008,
    "highq": 1808,
    "hihatcl": 4560,
    "hihatop": 20048,
    "hihatpd": 1808,
    "hitbrass": 31520,
    "homeorg": 992,
    "honky": 65680,
    "hrpschrd": 3584,
    "jazzgtr": 27712,
    "jingles": 16944,
    "jungle": 13616,
    "kalimba": 2208,
    "kick1": 4544,
    "kick2": 5024,
    "koto": 20832,
    "lead5th": 6464,
    "maracas": 4560,
    "marcato": 61232,
    "marimba": 2064,
    "metalpad": 30288,
    "metbell": 112,
    "metclick": 112,
    "musicbox": 15312,
    "mutegtr": 17008,
    "mutetrum": 9168,
    "nyguitar": 19200,
    "oboe": 3952,
    "ocarina": 1616,
    "odguitar": 12640,
    "orchhit": 14208,
    "percorg": 7520,
    "piccolo": 4320,
    "pickbass": 16416,
    "pistol": 18144,
    "pizzcato": 19888,
    "polysyn": 30224,
    "recorder": 2656,
    "reedorg": 1568,
    "revcym": 13536,
    "rockorg": 30288,
    "santur": 21760,
    "sawwave": 27056,
    "scratch1": 4384,
    "scratch2": 2288,
    "seashore": 31040,
    "shakazul": 31136,
    "shaker": 3104,
    "shamisen": 13136,
    "shannai": 9792,
    "sitar": 18288,
    "slap": 5856,
    "slapbas1": 27872,
    "slapbas2": 20592,
    "slowstr": 18192,
    "snare1": 8544,
    "snare2": 4096,
    "soundtrk": 19888,
    "sprnosax": 7072,
    "sqrclick": 112,
    "sqrwave": 15056,
    "startrak": 27376,
    "steeldrm": 11952,
    "stickrim": 2848,
    "sticks": 4224,
    "surdo1": 9600,
    "surdo2": 9600,
    "sweeper": 31216,
    "synbass1": 6160,
    "synbass2": 2928,
    "synbras1": 30704,
    "synbras2": 30160,
    "synpiano": 5456,
    "synstr1": 31216,
    "synstr2": 16416,
    "syntom": 30512,
    "taiko": 18672,
    "tamborin": 8944,
    "telephon": 4416,
    "tenorsax": 8448,
    "timbaleh": 5264,
    "timbalel": 9728,
    "timpani": 7072,
    "tomhi1": 6576,
    "tomhi2": 6560,
    "tomlo1": 6560,
    "tomlo2": 9600,
    "tommid1": 6560,
    "tommid2": 6560,
    "toms": 6576,
    "tremstr": 61232,
    "triangl1": 2224,
    "triangl2": 15792,
    "trombone": 12896,
    "trumpet": 6608,
    "tuba": 5760,
    "tubebell": 9120,
    "unicorn": 30096,
    "vibes": 10640,
    "vibslap": 9456,
    "viola": 27952,
    "violin": 12160,
    "voices": 14976,
    "voxlead": 14992,
    "warmpad": 18080,
    "whistle": 5872,
    "whistle1": 2000,
    "whistle2": 928,
    "woodblk": 3680,
    "woodblk1": 2352,
    "woodblk2": 3680,
    "woodflut": 1936,
    "xylophon": 9376,
}

# Groups of "similar sounding" instruments. The first instrument in each
# group is the "leader" and will be used as the fallback for other
# instruments in the group if they are not popular enough to be included.
#
# These groups are based on having listened to the instruments in the
# GUS patch set using the generated comparison mid (see comparison.py),
# with similar sounding instruments being grouped together.
#
# If you want to improve the generated config, here's where to start.
# Separating out into more, smaller groups helps, but the 256KB
# config's limited size is quite restrictive. In particular, it's
# important that the "leader" instrument for each group is small
# (see table above of patch sizes).

SIMILAR_GROUPS = [
    # Pianos.
    ("synpiano", "acpiano", "britepno", "honky"),
    ("glocken", "epiano1", "epiano2", "celeste"),
    # Harpsichord sounds noticeably different to pianos:
    ("hrpschrd", "clavinet"),
    # Xylophone etc.
    (
        "marimba",
        "musicbox",
        "vibes",
        "xylophon",
        "tubebell",
        "carillon",
        "santur",
        "kalimba",
    ),
    # Organs.
    ("homeorg", "percorg", "rockorg", "church", "reedorg"),
    # Accordion/Harmonica:
    ("accordn", "harmonca", "concrtna"),
    # Guitars.
    ("nyguitar", "acguitar", "jazzgtr"),
    # Overdriven/distortion guitars sound different. Besides, we
    # definitely want at least one of these.
    ("odguitar", "distgtr", "gtrharm", "cleangtr", "bagpipes"),
    # Basses.
    ("synbass2", "acbass", "fngrbass", "pickbass", "basslead", "fretless"),
    ("synbass1", "slapbas1", "slapbas2", "mutegtr"),
    # Violin and similar string instruments.
    ("violin", "viola", "cello", "contraba", "pizzcato", "harp"),
    # Other stringed (?)
    (
        "synstr2",
        "slowstr",
        "marcato",
        "synstr1",
        "choir",
        "doo",
        "voices",
        "orchhit",
        "polysyn",
        "bowglass",
        "tremstr",
        "fantasia",
        "warmpad",
        "ghostie",
        "metalpad",
        "sweeper",
        "halopad",
    ),
    # Trumpet and other brass.
    (
        "trumpet",
        "trombone",
        "tuba",
        "mutetrum",
        "frenchrn",
        "hitbrass",
        "synbras1",
        "synbras2",
    ),
    # Reed instruments.
    (
        "altosax",
        "sprnosax",
        "tenorsax",
        "barisax",
        "oboe",
        "englhorn",
        "bassoon",
        "clarinet",
    ),
    # Pipe instruments.
    (
        "recorder",
        "flute",
        "piccolo",
        "woodflut",
        "bottle",
        "shakazul",
        "whistle",
        "ocarina",
        "fiddle",
        "shannai",
        "calliope",
        "chiflead",
        "charang",
    ),
    # Leads:
    ("sqrwave", "sawwave", "voxlead", "lead5th"),
    # Odd stringed instruments.
    ("sitar", "banjo", "shamisen", "koto"),
    # Percussion sounds.
    # Kick:
    ("kick2", "taiko", "kick1"),
    # Conga:
    ("congahi2", "congahi1", "congalo"),
    # Snare drums:
    ("snare2", "claps", "snare1"),
    # Toms:
    (
        "tomlo1",
        "toms",
        "syntom",
        "tomlo2",
        "tommid1",
        "tommid2",
        "tomhi2",
        "tomhi1",
        "timpani",
    ),
    # Cymbal crash:
    ("cymsplsh", "cymcrsh2", "cymcrsh1", "revcym", "cymchina"),
    # Cymbal ride:
    ("cymride1", "cymride2", "cymbell", "hihatop"),
    # Hi-hat:
    ("hihatpd", "hihatcl"),
    # Metallic sounding:
    ("bongohi", "bongolo", "timbaleh", "timbalel", "cowbell"),
    # Click:
    ("stickrim", "woodblk1", "woodblk2", "woodblk", "clave"),
    # Random instruments we don't include unless they're popular enough.
    (
        "blank",
        # Special effects:
        "unicorn",
        "soundtrk",
        "aurora",
        "crystal",
        "atmosphr",
        "freshair",
        "echovox",
        "startrak",
        "fx-fret",
        "fx-blow",
        "seashore",
        "jungle",
        "telephon",
        "helicptr",
        "applause",
        "pistol",
        # Percussion:
        "cabasa",
        "whistle1",
        "whistle2",
        "vibslap",
        "maracas",
        "guiro1",
        "guiro2",
        "cuica1",
        "cuica2",
        "steeldrm",
        "agogohi",
        "agogolo",
        "agogo",
        "triangl1",
        "triangl2",
        "tamborin",
    ),
]



# SPDX-License-Identifier: BSD-3-Clause
#
# Generate comparison MIDI file.
#
# The comparison MIDI is used for testing and tweaking the similarity
# groups in the configuration file. In each group, the instruments in
# the group should sound broadly similar, and the first in the group
# should be able to substitute for any member of the group.
#
# Each similarity group is played in order, with a pause between each
# instrument. A programme of the instruments is printed to stdout so
# that the user can identify each instrument in order.
#
# To use the comparison MIDI, use a port that supports Timidity for
# MIDI playback, and configure it to use the GUS patch set (use
# music/dgguspat.zip from the idgames archive). Use the "full"
# configuration file (timidity.cfg in the .zip) so that every
# instrument is used for playback. Rename comparison.mid to
# d_runnin.lmp and run with '-file d_runnin.lmp -warp 1', so that the
# music plays.


from config import *
import midi


def instrument_num(instrument):
    """Given a GUS patch name, get the MIDI instrument number.

       For percussion instruments, the instrument number is offset
       by 128.
    """
    for key, name in GUS_INSTR_PATCHES.items():
        if name == instrument:
            return key
    raise Exception("Unknown instrument %s" % instrument)


pattern = midi.Pattern(resolution=48)
track = midi.Track()
track.append(midi.ControlChangeEvent(tick=0, channel=9, data=[7, 92]))

time = 500
for group in SIMILAR_GROUPS:
    # Don't bother with special effects.
    # if group[0] == 'blank':
    #   continue

    print("Group: ")
    empty = True
    for instrument in group:
        inum = instrument_num(instrument)

        # For normal instruments, play a couple of different notes.
        # For percussion instruments, play a couple of note on
        # the percussion channel.
        if inum <= 128:
            print("\t%s (%i)" % (instrument, inum))
            track.extend(
                [
                    midi.ProgramChangeEvent(tick=time, channel=0, data=[inum]),
                    midi.NoteOnEvent(
                        tick=0, channel=0, velocity=92, pitch=midi.A_3
                    ),
                    midi.NoteOffEvent(
                        tick=50, channel=0, velocity=92, pitch=midi.A_3
                    ),
                    midi.NoteOnEvent(
                        tick=0, channel=0, velocity=92, pitch=midi.B_3
                    ),
                    midi.NoteOffEvent(
                        tick=50, channel=0, velocity=92, pitch=midi.B_3
                    ),
                ]
            )
        else:
            print("\t%s (percussion %i)" % (instrument, inum - 128))
            track.extend(
                [
                    midi.NoteOnEvent(
                        tick=time, channel=9, velocity=92, pitch=inum - 128
                    ),
                    midi.NoteOffEvent(tick=50, channel=9, pitch=inum - 128),
                    midi.NoteOnEvent(
                        tick=0, channel=9, velocity=92, pitch=inum - 128
                    ),
                    midi.NoteOffEvent(tick=50, channel=9, pitch=inum - 128),
                ]
            )

        empty = False
        time = 100

    if not empty:
        time = 500

# Four drumbeats indicate the end of the track and that the music is
# going to loop.
for i in range(4):
    track.extend(
        [
            midi.NoteOnEvent(tick=20, channel=9, velocity=92, pitch=35),
            midi.NoteOffEvent(tick=20, channel=9, pitch=35),
        ]
    )

track.append(midi.EndOfTrackEvent(tick=1))
pattern.append(track)

# Save the pattern to disk
midi.write_midifile("comparison.mid", pattern)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause

import re

from PIL import Image


def get_image_dimensions(filename):
    """Get image dimensions w x h

    Args:
      filename: filename of the image
    """
    with Image.open(filename) as img:
        width, height = img.size
    return (width, height)


if __name__ == "__main__":
    import sys

    x, y = get_image_dimensions(sys.argv[1])
    string = "%i %i" % (x, y)
    sys.stdout.write(string)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Module to load Creative SBI OPL instrument files.
#

import struct
import sys

HEADER_VALUE = "SBI\x1a"

FIELDS = [
    "m_am_vibrato_eg",
    "c_am_vibrato_eg",
    "m_ksl_volume",
    "c_ksl_volume",
    "m_attack_decay",
    "c_attack_decay",
    "m_sustain_release",
    "c_sustain_release",
    "m_waveform",
    "c_waveform",
    "feedback_fm",
]


def read(filename):
    with open(filename, "rb") as f:
        data = f.read()

    header, name = struct.unpack("4s32s", data[0:36])
    header = header.decode("ascii")

    if header != HEADER_VALUE:
        raise Exception("Invalid header for SBI file!")

    instr_data = data[36:]
    result = {"name": name.decode("ascii").rstrip("\0")}

    for i in range(len(FIELDS)):
        result[FIELDS[i]], = struct.unpack("B", instr_data[i : i + 1])

    return result


def write(filename, data):
    with open(filename, "wb") as f:
        f.write(struct.pack("4s", HEADER_VALUE.encode("ascii")))
        f.write(struct.pack("32s", data["name"].encode("ascii")))

        for field in FIELDS:
            f.write(struct.pack("B", data[field]))
        for x in range(16 - len(FIELDS)):
            f.write(struct.pack("B", 0))


if __name__ == "__main__":
    for filename in sys.argv[1:]:
        print(filename)
        print(read(filename))



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
#
# Config file for GENMIDI file generation.
#
# This specifies the instrument files that should be used to generate
# the GENMIDI lump. Instrument files can be in SBI or A2I format.
# Instrument files must not use OPL3 features and it is an error to do so.
#
# A simple instrument specification looks like this:
#
#   Instrument("filename.sbi")
#
# For a double-voice instrument (two instrument files played at the same
# time), do this:
#
#   Instrument("file1.sbi", "file2.sbi")
#
# To tune the instruments, it's possible to apply an offset to the notes
# that are played. For example, to force all notes down by one octave:
#
#   Instrument("file1.sbi", off1=-12)
#
# This can be controlled individually for double-voice instruments, eg.
#
#   Instrument("file1.sbi", "file2.sbi", off1=-12, off2=+6)
#
# When an instrument is played, the note usually comes from the MIDI
# event. Some instruments (especially percussion) always play the same
# note. To specify a fixed note, do this:
#
#   Instrument("filename.sbi", note=On5.B)
#
# Value is a MIDI note number; see midi.py for constant definitions.

from instrument import Instrument, NullInstrument
from midi import *

# General MIDI instruments:

INSTRUMENTS = [
    Instrument("instr001.sbi"),  # #001 - Acoustic Grand Piano
    Instrument("instr002.sbi"),  # #002 - Bright Acoustic Piano
    Instrument("instr003.sbi"),  # #003 - Electric Grand Piano
    Instrument("instr004.sbi", "instr004-2.sbi"),  # #004 - Honky-tonk Piano
    Instrument("instr005.sbi", "instr005-2.sbi"),  # #005 - Electric Piano 1
    Instrument("instr006.sbi", off1=-12),  # #006 - Electric Piano 2
    Instrument("instr007.sbi"),  # #007 - Harpsichord
    Instrument("instr008.sbi", off1=-12),  # #008 - Clavi
    Instrument("instr009.sbi", off1=-12),  # #009 - Celesta
    Instrument("instr010.sbi", off1=-12),  # #010 - Glockenspiel
    Instrument("instr011.sbi", off1=-12),  # #011 - Music Box
    Instrument("instr012.sbi", off1=-12),  # #012 - Vibraphone
    Instrument("instr013.sbi", off1=-12),  # #013 - Marimba
    Instrument("instr014.sbi", off1=-12),  # #014 - Xylophone
    Instrument("instr015.sbi", off1=-12),  # #015 - Tubular Bells
    Instrument("instr016.sbi", off1=-12),  # #016 - Dulcimer
    Instrument("instr017.sbi", off1=-12),  # #017 - Drawbar Organ
    Instrument("instr018.sbi", off1=-12),  # #018 - Percussive Organ
    Instrument("instr019.sbi", off1=-12),  # #019 - Rock Organ
    Instrument("instr020.sbi", off1=-12),  # #020 - Church Organ
    Instrument("instr021.sbi", off1=-12),  # #021 - Reed Organ
    Instrument("instr022.sbi", off1=-12),  # #022 - Accordion
    Instrument("instr023.sbi", off1=-12),  # #023 - Harmonica
    Instrument("instr024.sbi", off1=-12),  # #024 - Tango Accordion
    Instrument("instr025.sbi", off1=-12),  # #025 - Acoustic Guitar (nylon)
    Instrument("instr026.sbi", off1=-12),  # #026 - Acoustic Guitar (steel)
    Instrument("instr027.sbi", off1=-12),  # #027 - Electric Guitar (jazz)
    Instrument("instr028.sbi", off1=-12),  # #028 - Electric Guitar (clean)
    Instrument("instr029.sbi", off1=-12),  # #029 - Electric Guitar (muted)
    Instrument("instr030.sbi", off1=-12),  # #030 - Overdriven Guitar
    Instrument("instr031.sbi", off1=-12),  # #031 - Distortion Guitar
    Instrument("instr032.sbi", off1=-12),  # #032 - Guitar harmonics
    Instrument("instr033.sbi", off1=-12),  # #033 - Acoustic Bass
    Instrument("instr034.sbi", off1=-12),  # #034 - Electric Bass (finger)
    Instrument("instr035.sbi", off1=-12),  # #035 - Electric Bass (pick)
    Instrument("instr036.sbi", off1=-12),  # #036 - Fretless Bass
    Instrument("instr037.sbi", off1=-12),  # #037 - Slap Bass 1
    Instrument("instr038.sbi", off1=-12),  # #038 - Slap Bass 2
    Instrument("instr039.sbi", off1=-12),  # #039 - Synth Bass 1
    Instrument("instr040.sbi", off1=-12),  # #040 - Synth Bass 2
    Instrument("instr041.sbi", off1=-12),  # #041 - Violin
    Instrument("instr042.sbi", off1=-12),  # #042 - Viola
    Instrument("instr043.sbi", off1=-12),  # #043 - Cello
    Instrument("instr044.sbi", off1=-12),  # #044 - Contrabass
    Instrument("instr045.sbi", off1=-12),  # #045 - Tremolo Strings
    Instrument("instr046.sbi", off1=-12),  # #046 - Pizzicato Strings
    Instrument("instr047.sbi", off1=-12),  # #047 - Orchestral Harp
    Instrument("instr048.sbi", off1=-12),  # #048 - Timpani
    Instrument("instr049.sbi", off1=-12),  # #049 - String Ensemble 1
    Instrument("instr050.sbi", off1=-12),  # #050 - String Ensemble 2
    Instrument("instr051.sbi", off1=-12),  # #051 - SynthStrings 1
    Instrument("instr052.sbi", off1=-12),  # #052 - SynthStrings 2
    Instrument("instr053.sbi", off1=-12),  # #053 - Choir Aahs
    Instrument("instr054.sbi", off1=-12),  # #054 - Voice Oohs
    Instrument("instr055.sbi", off1=-12),  # #055 - Synth Voice
    Instrument("instr056.sbi", off1=-12),  # #056 - Orchestra Hit
    Instrument("instr057.sbi", off1=-12),  # #057 - Trumpet
    Instrument("instr058.sbi", off1=-12),  # #058 - Trombone
    Instrument("instr059.sbi", off1=-12),  # #059 - Tuba
    Instrument("instr060.sbi", off1=-12),  # #060 - Muted Trumpet
    Instrument("instr061.sbi", off1=-12),  # #061 - French Horn
    Instrument("instr062.sbi", off1=-12),  # #062 - Brass Section
    Instrument("instr063.sbi", off1=-12),  # #063 - SynthBrass 1
    Instrument("instr064.sbi", off1=-12),  # #064 - SynthBrass 2
    Instrument("instr065.sbi", off1=-12),  # #065 - Soprano Sax
    Instrument("instr066.sbi", off1=-12),  # #066 - Alto Sax
    Instrument("instr067.sbi", off1=-12),  # #067 - Tenor Sax
    Instrument("instr068.sbi", off1=-12),  # #068 - Baritone Sax
    Instrument("instr069.sbi", off1=-12),  # #069 - Oboe
    Instrument("instr070.sbi", off1=-12),  # #070 - English Horn
    Instrument("instr071.sbi", off1=-12),  # #071 - Bassoon
    Instrument("instr072.sbi", off1=-12),  # #072 - Clarinet
    Instrument("instr073.sbi", off1=-12),  # #073 - Piccolo
    Instrument("instr074.sbi", off1=-12),  # #074 - Flute
    Instrument("instr075.sbi", off1=-12),  # #075 - Recorder
    Instrument("instr076.sbi", off1=-12),  # #076 - Pan Flute
    Instrument("instr077.sbi", off1=-12),  # #077 - Blown Bottle
    Instrument("instr078.sbi", off1=-12),  # #078 - Shakuhachi
    Instrument("instr079.sbi", off1=-12),  # #079 - Whistle
    Instrument("instr080.sbi", off1=-12),  # #080 - Ocarina
    Instrument("instr081.sbi", off1=-12),  # #081 - Lead 1 (square)
    Instrument("instr082.sbi"),  # #082 - Lead 2 (sawtooth)
    Instrument("instr083.sbi", off1=-12),  # #083 - Lead 3 (calliope)
    Instrument("instr084.sbi", off1=-12),  # #084 - Lead 4 (chiff)
    Instrument("instr085.sbi", off1=-12),  # #085 - Lead 5 (charang)
    Instrument("instr086.sbi", off1=-12),  # #086 - Lead 6 (voice)
    Instrument("instr087.sbi", off1=-12),  # #087 - Lead 7 (fifths)
    Instrument("instr088.sbi", off1=-12),  # #088 - Lead 8 (bass + lead)
    Instrument("instr089.sbi", off1=-12),  # #089 - Pad 1 (new age)
    Instrument("instr090.sbi", off1=-12),  # #090 - Pad 2 (warm)
    Instrument("instr091.sbi", off1=-12),  # #091 - Pad 3 (polysynth)
    Instrument("instr092.sbi", off1=-12),  # #092 - Pad 4 (choir)
    Instrument("instr093.sbi", off1=-12),  # #093 - Pad 5 (bowed)
    Instrument("instr094.sbi", off1=-12),  # #094 - Pad 6 (metallic)
    Instrument("instr095.sbi", off1=-12),  # #095 - Pad 7 (halo)
    Instrument("instr096.sbi", off1=-12),  # #096 - Pad 8 (sweep)
    Instrument("instr097.sbi", off1=-12),  # #097 - FX 1 (rain)
    Instrument("instr098.sbi", off1=-12),  # #098 - FX 2 (soundtrack)
    Instrument("instr099.sbi", off1=-12),  # #099 - FX 3 (crystal)
    Instrument("instr100.sbi", off1=-12),  # #100 - FX 4 (atmosphere)
    Instrument("instr101.sbi", off1=-12),  # #101 - FX 5 (brightness)
    Instrument("instr102.sbi", off1=-12),  # #102 - FX 6 (goblins)
    Instrument("instr103.sbi", off1=-12),  # #103 - FX 7 (echoes)
    Instrument("instr104.sbi", off1=-12),  # #104 - FX 8 (sci-fi)
    Instrument("instr105.sbi", off1=-12),  # #105 - Sitar
    Instrument("instr106.sbi", off1=-12),  # #106 - Banjo
    Instrument("instr107.sbi", off1=-12),  # #107 - Shamisen
    Instrument("instr108.sbi", off1=-12),  # #108 - Koto
    Instrument("instr109.sbi", off1=-12),  # #109 - Kalimba
    Instrument("instr110.sbi", off1=-12),  # #110 - Bag pipe
    Instrument("instr111.sbi", off1=-12),  # #111 - Fiddle
    Instrument("instr112.sbi", off1=-12),  # #112 - Shanai
    Instrument("instr113.sbi", off1=-12),  # #113 - Tinkle Bell
    Instrument("instr114.sbi", off1=-12),  # #114 - Agogo
    Instrument("instr115.sbi", off1=-12),  # #115 - Steel Drums
    Instrument("instr116.sbi", off1=-12),  # #116 - Woodblock
    Instrument("instr117.sbi", off1=-12),  # #117 - Taiko Drum
    Instrument("instr118.sbi", off1=-12),  # #118 - Melodic Tom
    Instrument("instr119.sbi", off1=-12),  # #119 - Synth Drum
    Instrument("instr120.sbi", off1=-12),  # #120 - Reverse Cymbal
    Instrument("instr121.sbi", off1=-12),  # #121 - Guitar Fret Noise
    Instrument("instr122.sbi", off1=-12),  # #122 - Breath Noise
    Instrument("instr123.sbi", off1=-12),  # #123 - Seashore
    Instrument("instr124.sbi", off1=-12),  # #124 - Bird Tweet
    Instrument("instr125.sbi", off1=-12),  # #125 - Telephone Ring
    Instrument("instr126.sbi", off1=-12),  # #126 - Helicopter
    Instrument("instr127.sbi", off1=-12),  # #127 - Applause
    Instrument("instr128.sbi", off1=-12),  # #128 - Gunshot
]

# Percussion instruments:
# Note: Many of these instruments have not been defined yet.
# As these are percussion instruments, a note number should ALWAYS be
# specified.

PERCUSSION = [
    Instrument("perc35.sbi", note=On4.A),  # #35 Acoustic Bass Drum
    Instrument("perc36.sbi", note=On4.A),  # #36 Bass Drum 1
    Instrument("perc37.sbi", note=On1.C),  # #37 Side Stick
    Instrument("perc38.sbi", note=On3.Gs),  # #38 Acoustic Snare
    Instrument("perc39.sbi", note=O3.C),  # #39 Hand Clap
    Instrument("perc40.sbi", note=On1.Cs),  # #40 Electric Snare
    Instrument("perc41.sbi", note=On3.D),  # #41 Low Floor Tom
    Instrument("perc42.sbi", note=O1.Gs),  # #42 Closed Hi Hat
    Instrument("perc43.sbi", note=On3.Gs),  # #43 High Floor Tom
    Instrument("perc44.sbi", note=O1.Gs),  # #44 Pedal Hi-Hat
    Instrument("perc45.sbi", note=On2.C),  # #45 Low Tom
    Instrument("perc46.sbi", note=O1.Gs),  # #46 Open Hi-Hat
    Instrument("perc47.sbi", note=On2.Fs),  # #47 Low-Mid Tom
    Instrument("perc48.sbi", note=On2.A),  # #48 Hi-Mid Tom
    Instrument("perc49.sbi", note=On1.C),  # #49 Crash Cymbal 1
    Instrument("perc50.sbi", note=On1.Cs),  # #50 High Tom
    Instrument("perc51.sbi", note=On1.B),  # #51 Ride Cymbal 1
    Instrument("perc52.sbi", note=On1.C),  # #52 Chinese Cymbal
    Instrument("perc53.sbi", note=O1.E),  # #53 Ride Bell
    Instrument("perc54.sbi", note=O0.E),  # #54 Tambourine
    NullInstrument,  # TODO - #55 Splash Cymbal
    NullInstrument,  # TODO - #56 Cowbell
    Instrument("perc57.sbi", note=On1.As),  # #57 Crash Cymbal 2
    NullInstrument,  # TODO - #58 Vibraslap
    Instrument("perc59.sbi", note=O0.E),  # #59 Ride Cymbal 2
    NullInstrument,  # TODO - #60 Hi Bongo
    NullInstrument,  # TODO - #61 Low Bongo
    NullInstrument,  # TODO - #62 Mute Hi Conga
    NullInstrument,  # TODO - #63 Open Hi Conga
    NullInstrument,  # TODO - #64 Low Conga
    NullInstrument,  # TODO - #65 High Timbale
    NullInstrument,  # TODO - #66 Low Timbale
    NullInstrument,  # TODO - #67 High Agogo
    NullInstrument,  # TODO - #68 Low Agogo
    NullInstrument,  # TODO - #69 Cabasa
    Instrument("perc70.sbi", note=On5.E),  # #70 Maracas
    Instrument("perc71.sbi", note=On5.E),  # #71 Short Whistle
    Instrument("perc72.sbi", note=On5.E),  # #72 Long Whistle
    Instrument("perc73.sbi", note=On5.E),  # #73 Short Guiro
    Instrument("perc74.sbi", note=On5.E),  # #74 Long Guiro
    Instrument("perc75.sbi", note=On5.E),  # #75 Claves
    Instrument("perc76.sbi", note=On5.E),  # #76 Hi Wood Block
    Instrument("perc77.sbi", note=On5.E),  # #77 Low Wood Block
    Instrument("perc78.sbi", note=On5.E),  # #78 Mute Cuica
    Instrument("perc79.sbi", note=On5.E),  # #79 Open Cuica
    Instrument("perc80.sbi", note=On5.E),  # #80 Mute Triangle
    Instrument("perc81.sbi", note=On5.E),  # #81 Open Triangle
]



# SPDX-License-Identifier: BSD-3-Clause
#
# Configuration file for textgen. This file defines the graphic lumps
# that are generated, and the text to show in each one.
#

import re

# Adjustments for character position based on character pairs. Some
# pairs of characters can fit more snugly together, which looks more
# visually appealing. This is highly dependent on the font graphics,
# and if the font is changed this probably needs to be redone.

# TODO: Add more rule for lower-case characters.

FONT_KERNING_RULES = {
    # Right character fits under left character:
    r"[TY][07ACOSZ]": -2,
    r"[TYty][a]": -2,
    r"P[A]": -3,
    r"P[7]": -2,
    r"P[Z]": -1,
    r"7[Z]": -1,
    r"[0OQ]A": -1,
    r"S[A]": -1,
    r"V[0OC]": -2,
    # Left character fits under right character:
    r"L[TY]": -4,
    r"L[014COQV]": -3,
    r"L[9]": -2,
    r"[0O][4TY]": -2,
    r"[0O][1]": -1,
    r"Q[1T]": -2,
    r"Q[Y]": -1,
    r"A[TYV]": -2,
    r"A[GC]": -1,
    r"a[TYty]": -2,
    r"a[vV]": -2,
    r"a[g]": -1,
    # Fits into "hole" in left character:
    r"[BCX8][0CGOQ]": -2,
    r"Z[0CO]": -2,
    r"Z[GQ]": -1,
    r"I[0COQ]": -1,
    r"K[0CO]": -4,
    r"K[GQ]": -3,
    r"K[E]": -1,
    r"[PR][0COQ]": -1,
    # Fits into "hole" in right character:
    r"[O0Q][X]": -3,
    r"[O0Q][28]": -2,
    r"[O0Q][9IK]": -1,
    # Just because.
    r"[O0][O0]": -1,
}

white_graphics = {
    "wibp1": "P1",
    "wibp2": "P2",
    "wibp3": "P3",
    "wibp4": "P4",
    "wicolon": ":",
    # These files are for the title screens of Phase 1 and Phase 2
    "t_phase1": "PHASE 1",
    "t_phase2": "PHASE 2",
    # Note: level names are also included in this dictionary, with
    # the data added programatically from the DEHACKED lump, see
    # code below.
}

blue_graphics = {
    "m_disopt": "DISPLAY OPTIONS",
    "m_episod": "Choose Chapter:",
    "m_optttl": "OPTIONS",
    "m_skill": "Choose Skill Level:",
}

red_graphics = {
    # Title for the HELP/HELP1 screen:
    "helpttl": "Help",
    # Title for CREDIT
    "freettl": "Freedoom",
    "m_ngame": "New Game",
    "m_option": "Options",
    "m_loadg": "Load Game",
    "m_saveg": "Save Game",
    "m_rdthis": "Read This!",
    "m_quitg": "Quit Game",
    "m_newg": "NEW GAME",
    "m_epi1": "Outpost Outbreak",
    "m_epi2": "Military Labs",
    "m_epi3": "Event Horizon",
    "m_epi4": "Double Impact",
    "m_jkill": "Please don't kill me!",
    "m_rough": "Will this hurt?",
    "m_hurt": "Bring on the pain.",
    "m_ultra": "Extreme Carnage.",
    "m_nmare": "MAYHEM!",
    "m_lgttl": "LOAD GAME",
    "m_sgttl": "SAVE GAME",
    "m_endgam": "End Game",
    "m_messg": "Messages:",
    "m_msgoff": "off",
    "m_msgon": "on",
    "m_msens": "Mouse Sensitivity",
    "m_detail": "Graphic Detail:",
    "m_gdhigh": "high",
    "m_gdlow": "low",
    "m_scrnsz": "Screen Size",
    "m_svol": "Sound Volume",
    "m_sfxvol": "Sfx Volume",
    "m_musvol": "Music Volume",
    "m_disp": "Display",
    "wif": "finished",
    "wiostk": "kills",
    "wiosti": "items",
    "wiscrt2": "secret",
    "wiosts": "scrt",
    "wifrgs": "frgs",
    "witime": "Time:",
    "wisucks": "sucks",
    "wimstt": "Total:",
    "wipar": "Par:",
    "wip1": "P1",
    "wip2": "P2",
    "wip3": "P3",
    "wip4": "P4",
    "wiostf": "f.",
    "wimstar": "you",
    "winum0": "0",
    "winum1": "1",
    "winum2": "2",
    "winum3": "3",
    "winum4": "4",
    "winum5": "5",
    "winum6": "6",
    "winum7": "7",
    "winum8": "8",
    "winum9": "9",
    "wipcnt": "%",
    "wiminus": "-",
    "wienter": "ENTERING",
    "m_pause": "pause",
    # Extra graphics used in PrBoom's menus. Generate these as well
    # so that when we play in PrBoom the menus look consistent.
    "prboom": "PrBoom",
    "m_generl": "General",
    "m_setup": "Setup",
    "m_keybnd": "Key Bindings",
    "m_weap": "Weapons",
    "m_stat": "Status Bar/HUD",
    "m_auto": "Automap",
    "m_enem": "Enemies",
    "m_mess": "Messages",
    "m_chat": "Chat Strings",
    "m_horsen": "horizontal",
    "m_versen": "vertical",
    "m_loksen": "mouse look",
    "m_accel": "acceleration",
    # Extra graphics from SMMU/Eternity Engine:
    "m_about": "about",
    "m_chatm": "Chat Strings",
    "m_compat": "Compatibility",
    "m_demos": "demos",
    "m_dmflag": "deathmatch flags",
    "m_etcopt": "eternity options",
    "m_feat": "Features",
    "m_gset": "game settings",
    "m_hud": "heads up display",
    "m_joyset": "joysticks",
    "m_ldsv": "Load/Save",
    "m_menus": "Menu Options",
    "m_mouse": "mouse options",
    "m_player": "player setup",
    "m_serial": "serial connection",
    "m_sound": "sound options",
    "m_status": "status bar",
    "m_tcpip": "tcp/ip connection",
    "m_video": "video options",
    "m_wad": "load wad",
    "m_wadopt": "wad options",
    # This is from SMMU too, and if we follow things to the letter,
    # ought to be all lower-case. However, same lump name is used
    # by other ports (Zandronum) which expect a taller graphic to
    # match the other main menu graphics. Eternity Engine doesn't
    # use it any more, and on SMMU there's enough space for it.
    "m_multi": "Multiplayer",
}


def read_bex_lump(filename):
    """Read the BEX (Dehacked) lump from the given filename.

    Returns:
        Dictionary mapping from name to value.
    """
    result = {}
    with open(filename) as f:
        for line in f:
            # Ignore comments:
            line = line.strip()
            if len(line) == 0 or line[0] in "#;":
                continue
            # Just split on '=' and interpret that as an
            # assignment. This is primitive and doesn't read
            # like a full BEX parser should, but it's good
            # enough for our purposes here.
            assign = line.split("=", 2)
            if len(assign) != 2:
                continue
            result[assign[0].strip()] = assign[1].strip()
    return result


def update_level_name(lumpname, bexdata, bexname):
    """Set the level name for the given graphic from BEX file.

    Args:
      lumpname: Name of output graphic file.
      bexdata: Dictionary of data read from BEX file.
      bexname: Name of entry in BEX file to use.
    """
    if bexname not in bexdata:
        raise Exception(
            "Level name %s not defined in " "DEHACKED lump!" % bexname
        )
    # Strip "MAP01: " or "E1M2: " etc. from start, if present:
    levelname = re.sub("^\w*\d:\s*", "", bexdata[bexname])
    white_graphics[lumpname] = levelname


freedoom_bex = read_bex_lump("../../lumps/p2_deh.lmp")
freedm_bex = read_bex_lump("../../lumps/fdm_deh.lmp")

for e in range(4):
    for m in range(9):
        # HUSTR_E1M1 from BEX => wilv00
        update_level_name(
            "wilv%i%i" % (e, m), freedoom_bex, "HUSTR_E%iM%i" % (e + 1, m + 1)
        )

for m in range(32):
    # HUSTR_1 => cwilv00
    update_level_name("cwilv%02i" % m, freedoom_bex, "HUSTR_%i" % (m + 1))
    # HUSTR_1 => dmwilv00 (from freedm.bex)
    update_level_name("dmwilv%02i" % m, freedm_bex, "HUSTR_%i" % (m + 1))



#!/usr/bin/env python3

# SPDX-License-Identifier: MIT
# Copyright (c) 2017 Martin Miller, Nick Zatkovich
# https://stackoverflow.com/questions/12251896/colorize-image-while-preserving-transparency-with-pil

from PIL import Image, ImageColor, ImageOps


def image_tint(image, tint=None):
    if tint is None:
        return image
    if image.mode not in ["RGB", "RGBA"]:
        image = image.convert("RGBA")

    tr, tg, tb = ImageColor.getrgb(tint)
    tl = ImageColor.getcolor(tint, "L")  # tint color's overall luminosity
    if not tl:
        tl = 1  # avoid division by zero
    tl = float(tl)  # compute luminosity preserving tint factors
    sr, sg, sb = map(
        lambda tv: tv / tl, (tr, tg, tb)
    )  # per component adjustments

    # create look-up tables to map luminosity to adjusted tint
    # (using floating-point math only to compute table)
    luts = (
        tuple(map(lambda lr: int(lr * sr + 0.5), range(256)))
        + tuple(map(lambda lg: int(lg * sg + 0.5), range(256)))
        + tuple(map(lambda lb: int(lb * sb + 0.5), range(256)))
    )
    l = ImageOps.grayscale(image)  # 8-bit luminosity version of whole image
    if Image.getmodebands(image.mode) < 4:
        merge_args = (image.mode, (l, l, l))  # for RGB verion of grayscale
    else:  # include copy of image's alpha layer
        a = Image.new("L", image.size)
        a.putdata(image.getdata(3))
        merge_args = (image.mode, (l, l, l, a))  # for RGBA verion of grayscale
        luts += tuple(range(256))  # for 1:1 mapping of copied alpha values

    return Image.merge(*merge_args).point(luts)


def main(input_image_path, tintcolor, result_image_path):
    image = Image.open(input_image_path)

    image.load()

    result = image_tint(image, tintcolor)
    if os.path.exists(result_image_path):  # delete any previous result file
        os.remove(result_image_path)
    result.save(result_image_path)  # file name's extension determines format


if __name__ == "__main__":
    import os
    import sys

    main(sys.argv[1], sys.argv[2], sys.argv[3])
